unit UCompile;

interface

uses
  Classes, SysUtils, WinProcs, Dialogs;

type
  TCompileThread = class(TThread)
  private
    { Private declarations }
  protected
    procedure Execute; override;
  end;
var
  Locator,
  InFileName,
  OutFileName,
  LogFileName           : String;
  InFile,
  OutFile,
  LogFile               : TextFile;

implementation

uses
  UMain, UFiles;


(******************************************************************************)


{ Important: Methods and properties of objects in VCL can only be used in a
  method called using Synchronize, for example,

      Synchronize(UpdateCaption);

  and UpdateCaption could look like,

    procedure TCompileThread.UpdateCaption;
    begin
      Form1.Caption := 'Updated in a thread';
    end; }

{ TCompileThread }

procedure TCompileThread.Execute;
  { Place thread code here }

{
Compile algorithm by J.W. Bizzaro, (c)1996-1998.

This program takes GenBank nucleotide documents (in the format used as of
this date) and compiles the sequences of a selected locator into a file.


1. Only for GenBank-formatted nucleotide documents

2. Place in directory with file containing GenBank documents.

3. No single document can have more than 2,000,000 base pairs.

4. All documents in this file will be examined.

5. The origin of the nucleotides can be specified, and they
   can be isolated to a higher degree of accuracy than with
   a simple database queue.

6. Specifying that a document contains more than one gene (as with
   chromosomes) allows flanking sequences to be identified when
   no intron sequences are found.

7. Partial sequences can be included or excluded.

8. Asterisks [*] can be added to mark the ends of sequences.

9. Two files will be produced:
   - One of lowercase nucleotide sequences (any letter a-z), with no line breaks
   - One that gives a log for the documents examined

10. Every nucleotide is tracked in the log file.  The names of the documents,
   the positions of the sequences in the documents, and their positions in
   the out files are all recorded in this file.

11. All existing files with these names will be overwritten without warning.
}

const
  SourceSizeLimit  = 2000000;
  Extension        = '.SEQ';

var
  DocsExaminedString,
  DocsCompiledString     : String;
  Code,
  ExaminedDocs,
  SkippedDocs            : Integer;
  InFileSize,
  CharactersRead,
  BasesExamined,
  BasesCompiled          : Real;
  Nucleus,
  Mitochon,
  Chloro,
  MultiGenes,
  PartialSeqs,
  MarkEnds               : Boolean;
  InFile,
  OutFile,
  LogFile                : TextFile;
  SearchRec              : TSearchRec;


(********************************************************************)

procedure WriteLog(LineOfText : String);

{ Procedure to consolidate the output of log to both LogFile and Memo }

begin

  Writeln(LogFile, LineOfText);
  FMain.Memo.Lines.Add(LineOfText);

end;


(********************************************************************)

procedure CompileAdoc;

{
Routine that searches the document for some keywords, the most
important of which being the one that identifies the sequence
location, known as the locator.  When the sequences themselves
are reached, the sequences are compiled.
}

type
  CodeLimit        = 0..SourceSizeLimit;
  LimitNotesType   = array[CodeLimit] of Boolean;

var
  Character        : Char;
  NextString,
  NextWord         : String;
  Y,
  Yremembered,
  SourceUpper,
  SeqLower,
  SeqUpper         : Integer;
  LimitNotes       : LimitNotesType;
  Done,
  FoundNucleus,
  FoundNucleusNucleo,
  FoundMitochon,
  FoundMitochonNucleo,
  FoundChloro,
  FoundChloroNucleo,
  FoundORGANISM,
  FoundREFERENCE,
  FoundLocator,
  FoundCDS,
  MultipleCDS,
  FoundExon,
  FoundIntron,
  FoundNumbers,
  MoreNumbers      : Boolean;


  (***********************************************)


  procedure ClearLimitNotes(Wanted : Boolean);

{
Routine that sets up the array that keeps track of which sequences are
wanted.
}

  var
    Z : Integer;

  begin { ClearLimitNotes }

    for Z := 0 to SourceSizeLimit do
      LimitNotes[Z] := Wanted;

  end; { ClearLimitNotes }


  (***********************************************)


  procedure ReadNextString(var NextString,
                               NextWord    : String);
{
Routine that forms groups of characters in the GenBank document into NextString
and groups of characters, excluding punctuation and converting UC to lc, into
NextWord.  The ends of both variables are where CR/LF and spaces are found.
}

  var
    ConvertedChar : Char;


    (*******************************************)


    procedure ConvertChar;
{ routine to get a lowercase letter a-z only }

    begin { ConvertChar }

    if Character in ['A'..'z']
      then
      begin
        if Character in ['A'..'Z']
          then
            ConvertedChar := Chr(Ord(Character) + 32)
          else
            ConvertedChar := Character;
      end;

    end; { ConvertChar }

    (*******************************************)


  begin  { ReadNextString }

    NextString := '';
    NextWord := '';
    Read(InFile, Character);
    CharactersRead := CharactersRead + 1;
    ConvertChar;

    { skip over CR/LF and spaces }
    while (not EOF(InFile)) and ((Ord(Character) = 10)
    or (Ord(Character) = 13) or (Character = ' ')) do
    begin
      Read(InFile, Character);
      CharactersRead := CharactersRead + 1;
      ConvertChar;
    end;

    { read in NextString while there are no CR/LF or spaces }
    while (not EOF(InFile)) and (not((Ord(Character) = 10)
    or (Ord(Character) = 13) or (Character = ' '))) do
    begin
      NextString := NextString + Character;

      if Character in ['A'..'z']
        then
          NextWord := NextWord + ConvertedChar;

      Read(InFile, Character);
      CharactersRead := CharactersRead + 1;
      ConvertChar;
    end;

  end; { ReadNextString }


  (***********************************************)


  procedure FindLimits(var Lower       : Integer;
                       var Upper       : Integer;
                       var MoreNumbers : Boolean);

{
Routine that is used after finding the locator that identifies where the
sequences lie.  It is assumed the following string is comprised of several
numbers that identify the location.  The numbers are used in the array that
keeps track of the locations of wanted sequences.
}

  label TryAgain;

  var
    LeftNumber,
    RightNumber      : String;

  begin { FindLimits }

    if MoreNumbers = True
      then
        Y := Yremembered
      else
      begin
        Y := 1;
        ReadNextString(NextString,NextWord);
      end;

    Yremembered := 1;
    MoreNumbers := False;

  TryAgain:

    LeftNumber := '';
    RightNumber := '';
    Lower := 0;
    Upper := 0;

    while not (NextString[Y] in ['0'..'9']) and (Y < Length(NextString)) do
    begin                                   { before numbers }
      Y := Y + 1;

      if (NextString[Y] = '<') and (PartialSeqs = False)
       and (Locator <> 'flanking')
        then
        begin
          WriteLog('Incomplete code found.');
          Done := True;
        end;

    end;

    while (NextString[Y] in ['0'..'9']) and (Y < Length(NextString)) do
    begin                                   { left number }
      LeftNumber := LeftNumber + NextString[Y];
      Y := Y + 1;
    end;

    while not (NextString[Y] in ['0'..'9']) and (Y < Length(NextString)) do
    begin                                   { between numbers }
      if (NextString[Y] = '>') and (PartialSeqs = False)
       and (Locator <> 'flanking')
        then
        begin
          WriteLog('Incomplete code found.');
          Done := True;
        end;

      { if a colon encountered here, the numbers apply to another document }
      if NextString[Y] = ':'
        then
        begin
          while (NextString[Y] <> ',') and (Y < Length(NextString)) do
            Y := Y + 1;

          if Y < Length(NextString)
            then
              goto tryAgain;
        end;

      Y := Y + 1;
    end;

    while (nextString[Y] in ['0'..'9']) and (Y <= Length(NextString)) do
    begin                                   { right number }
      RightNumber := RightNumber + NextString[Y];
      Y := Y + 1;
    end;

    while not (NextString[Y] in ['0'..'9']) and (Y < Length(NextString)) do
    begin                                   { after numbers }
      Y := Y + 1;
    end;

    if (NextString[Y] in ['0'..'9']) and (Y < Length(NextString))
      then                                  { if more numbers }
      begin
        MoreNumbers := True;
        Yremembered := Y;
      end;

    if (NextString[Y] = ',') and (Y = Length(NextString))
      then                                  { if line breaks }
      begin
        MoreNumbers := True;
        Y := 1;
        ReadNextString(NextString,NextWord);
      end;

    Val(LeftNumber, Lower, Code);
    Val(RightNumber, Upper, Code);

    { reset to 0 if anything is unexpected }
    if (Lower > Upper) or (Lower = 0) or (Upper = 0)
      then
      begin
        Lower := 0;
        Upper := 0;
      end
      else
        FoundNumbers := True;

  end; { FindLimits }


  (***********************************************)


  procedure NoteLimits(Wanted   : Boolean;
                       Lower,
                       Upper    : Integer);

{
Routine that stores into an array the locations of wanted sequences
in the GenBank document.
}

  var
    Z : Integer;

  begin { NoteLimits }

    for Z := Lower to Upper do
      LimitNotes[Z] := Wanted;

  end; { NoteLimits }


  (***********************************************)


  procedure CompileSequences;

{
Routine that compares the nucleotides in the GenBank document to the
information about them that was gathered and stored in an array (whether
they are wanted or not), and then it writes the nucleotides into the outfile.

start:   The wanted sequence starts here.
in:      This is one of the wanted bases.
stop:    The wanted sequence stops here.
out:     This is one of the unwanted bases.
}

  var
    Z      : Integer;
    Status : String;


    (*******************************************)


    procedure WriteLimitNotes(Z: integer);

{
Routine that writes the information that is known about the wanted sequences:
where in the GenBank document it came from and where in the outfile it's
written.  This information is written in the notes file.
}

    begin { WriteLimitNotes }

      if (Status = 'Start') or (Status = 'Stop')
        then
        begin

          WriteLog(Status + Chr(9) + Chr(9) + IntToStr(Z) + Chr(9) + Chr(9) + FloatToStrF(BasesCompiled,ffFixed,1,0));

        end; { if }

    end; { WriteLimitNotes }


    (*******************************************)


  begin { CompileSequences }

    Z := 1;
    Status := 'Out';

    Read(InFile, Character);
    CharactersRead := CharactersRead + 1;

    while (Z <= SourceUpper) and (Character <> '/') and (not EOF(InFile)) do
    begin

      if Character in ['a'..'z']
        then
        begin
          if (LimitNotes[Z] = False)
            then
            begin
              if Status = 'Stop'
                then
                  Status := 'Out'
            end
            else
            begin

              if Status = 'Start'
                then
                  Status := 'In';

              if (Status = 'Stop') or (Status = 'Out')
                then
                  Status := 'Start';

              if (LimitNotes[Z+1] = False) or (Z = SourceUpper)
                then
                  Status := 'Stop';

              Write(OutFile, Character);
              BasesCompiled := BasesCompiled + 1;

            end; { if }

            if (Status = 'Stop') and MarkEnds
              then
                Write(OutFile, '*'); { indication of where sequence ends }

            WriteLimitNotes(Z);
            Z := Z + 1;

        end; { if }

      Read(InFile, Character);
      CharactersRead := CharactersRead + 1;

    end; { while }

  end; { CompileSequences }


  (***********************************************)


begin { CompileAdoc }

  if Locator = 'flanking'
    then
      ClearLimitNotes(True)
    else
      ClearLimitNotes(False);

  Done := False;
  FoundNucleus := False;
  FoundNucleusNucleo := False;
  FoundMitochon := False;
  FoundMitochonNucleo := False;
  FoundChloro := False;
  FoundChloroNucleo := False;
  FoundORGANISM  := False;
  FoundREFERENCE := False;
  FoundLocator := False;
  FoundCDS := False;
  MultipleCDS := False;
  FoundExon := False;
  FoundIntron := False;

  ReadNextString(NextString,NextWord);

  while (Done = False) and (not EOF(InFile)) do
  begin
    MoreNumbers := False;
    Yremembered := 1;

    if NextString = 'LOCUS'
      then
      begin

        ExaminedDocs := ExaminedDocs + 1;

        WriteLog('');
        ReadNextString(NextString,NextWord);
        WriteLog(IntToStr(ExaminedDocs) + ': ' + NextString);

        ReadNextString(NextString,NextWord);
        Val(NextString, SourceUpper, Code);
        WriteLog(IntToStr(SourceUpper) + 'bp');
        BasesExamined := BasesExamined + SourceUpper;

        if SourceUpper > SourceSizeLimit
          then
          begin
            WriteLog('Too many base pairs in this document.');
            Done := True;
          end;

      end;

    if NextString = 'ORGANISM'
      then
        FoundORGANISM := True;

    if NextString = 'REFERENCE'
      then
        FoundREFERENCE := True;

    { identify nucleotide origin by keywords in document }
    { NextWord is needed because the keywords are normally found among symbols }
    if (NextWord = 'nuclear') or (NextWord = 'nucleus')
     or (NextWord = 'nuclei')
      then
      begin
        FoundNucleus := True;
        ReadNextString(NextString,NextWord);
        if (NextWord = 'dna') or (NextWord = 'rna') or (NextWord = 'mrna')
         or (NextWord = 'gene') or (NextWord = 'genes')
         or (NextWord = 'genome') or (NextWord = 'genomes')
         or (NextWord = 'nucleotide') or (NextWord = 'nucleotides')
          then FoundNucleusNucleo := True;
      end;

    if ((NextWord = 'mitochondrial') or (NextWord = 'mitochondrion')
     or (NextWord = 'mitochondria')) and (not (FoundORGANISM and (not FoundREFERENCE)))
      then                                   { mitochondri... often in taxonomy }
      begin
        FoundMitochon := True;
        ReadNextString(NextString,NextWord);
        if (NextWord = 'dna') or (NextWord = 'rna') or (NextWord = 'mrna')
         or (NextWord = 'gene') or (NextWord = 'genes')
         or (NextWord = 'genome') or (NextWord = 'genomes')
         or (NextWord = 'nucleotide') or (NextWord = 'nucleotides')
          then FoundMitochonNucleo := True;
      end;

    if ((NextWord = 'chloroplast') or (NextWord = 'chloroplasts'))
     and (not (FoundORGANISM and (not FoundREFERENCE)))
      then           { chloroplast often in taxonomy }
      begin
        FoundChloro := True;
        ReadNextString(NextString,NextWord);
        if (NextWord = 'dna') or (NextWord = 'rna') or (NextWord = 'mrna')
         or (NextWord = 'gene') or (NextWord = 'genes')
         or (NextWord = 'genome') or (NextWord = 'genomes')
         or (NextWord = 'nucleotide') or (NextWord = 'nucleotides')
          then FoundChloroNucleo := True;
      end;

    if NextString = Locator
      then
      begin
        FoundNumbers := False;
        FindLimits(SeqLower, SeqUpper, MoreNumbers);
        NoteLimits(True, SeqLower, SeqUpper);

        while MoreNumbers = True do
        begin
          FindLimits(SeqLower, SeqUpper, MoreNumbers);
          NoteLimits(True, SeqLower, SeqUpper)
        end;

        if FoundNumbers
          then
            FoundLocator := True;
      end;

    { exclude coding and intron sequences from what would be considered flanking }
    if (Locator = 'flanking') and ((NextString = 'CDS')
     or (NextString = 'exon') or (NextString = 'intron'))
      then
      begin
        if NextString = 'CDS'
          then
          begin
            if FoundCDS = True
              then
                MultipleCDS := True;
            FoundCDS := True;
          end;
        if NextString = 'exon'
          then
            FoundExon := True;
        if NextString = 'intron'
          then
            FoundIntron := True;

        FoundNumbers := False;
        FindLimits(SeqLower, SeqUpper, MoreNumbers);
        NoteLimits(False, SeqLower, SeqUpper);

        if (NextString = 'CDS') and (MoreNumbers = True)
          then
            MultipleCDS := True;

        while MoreNumbers = True do
        begin
          FindLimits(SeqLower, SeqUpper, MoreNumbers);
          NoteLimits(False, SeqLower, SeqUpper)
        end;

        Done := False; { Exempt from rule for incompletes. }

      end; { if }

    { ready to read in bases when 'ORIGIN' found }
    { consider whether or not the sequences from the document should be used }
    if NextString = 'ORIGIN'
      then
      begin

        { advance to start of sequences }
        while (NextString <> '1') and (not EOF(InFile)) do
          ReadNextString(NextString,NextWord);

        { check that the unwanted nucleotide origins are not present }
        { MitochonNucleo and ChloroNucleo terms override any nucleus terms }
        { unwanted FoundNucleus but no Mitochon or Chloro ...Nucleo }
        if (not Nucleus) and FoundNucleus and (not FoundMitochonNucleo)
         and (not FoundChloroNucleo)
          then
          begin
            WriteLog('Unwanted nucleus nucleotides probably present.');
            Done := True;
          end;

        { unwanted FoundMitochon alone or with FoundChloro but no FoundNucleus
          and no ...Nucleo }
        { FoundNucleus will override }
        if (not Mitochon) and FoundMitochon and (not FoundNucleus)
         and (not FoundMitochonNucleo) and (not FoundChloroNucleo)
          then
          begin
            WriteLog('Uncertain nucleotide origin, possibly of unwanted mitochondrion.');
            Done := True;
          end;

        { unwanted MitochonNucleo found }
        if (not Mitochon) and FoundMitochonNucleo
          then
          begin
            WriteLog('Unwanted mitochondrial nucleotides probably present.');
            Done := True;
          end;

        { unwanted FoundChloro alone or with FoundMitochon but no FoundNucleus
          and no ...Nucleo }
        { FoundNucleus will override }
        if (not Chloro) and FoundChloro and (not FoundNucleus)
         and (not FoundMitochonNucleo) and (not FoundChloroNucleo)
          then
          begin
            WriteLog('Uncertain nucleotide origin, possibly of unwanted chloroplast.');
            Done := True;
          end;

        { unwanted ChloroNucleo found }
        if (not Chloro) and FoundChloroNucleo
          then
          begin
            WriteLog('Unwanted chloroplast nucleotides probably present.');
            Done := True;
          end;

        { check that the document actually contains mitochondrial or chloroplast
          nucleotides if those are wanted }
        { if the document does not indicate what type it contains, it's assumed
          to contain nuclear nucleotides }
        { if FoundMitochon or FoundChloro but neither is ...Nucleo, then the
          document may contain nuclear, but it's uncertain }
        if (Mitochon and (not Nucleus) and (not Chloro) and (not FoundMitochonNucleo))
         or (Chloro and (not Nucleus) and (not Mitochon) and (not FoundChloroNucleo))
         or (Mitochon and Chloro and (not Nucleus)
         and (not (FoundMitochonNucleo or FoundChloroNucleo)))
          then
          begin
            WriteLog('Nucleotides of wanted origin probably not present.');
            Done := True;
          end;

        if (Locator <> 'flanking') and (not FoundLocator)
          then
          begin
            WriteLog('Locator not found.');
            Done := True;
          end;

        { verify if what is thought to be flanking really is }
        if (Locator = 'flanking') and (FoundLocator = False)
         and ((FoundCDS = False) or ((MultipleCDS = True)
         and (not MultiGenes) and (FoundIntron = False)))
          then
          begin
            WriteLog('Locations of flanking sequences uncertain.');
            Done := True;
          end;

        if Done = False
          then
          begin
            if MarkEnds
              then
                WriteLog('Block #' + IntToStr(ExaminedDocs - SkippedDocs));

            WriteLog(Chr(9) + Chr(9) + 'Start/Stop' + Chr(9) + 'Start/Stop');
            WriteLog(Chr(9) + Chr(9) + 'Source File' + Chr(9) + 'Dest. File');
            WriteLog(Chr(9) + Chr(9) + '----------' + Chr(9) + '----------');

            CompileSequences;
          end
          else
          begin
            WriteLog('This GenBank document was skipped.');
            SkippedDocs := SkippedDocs + 1;
          end;

        Done := True; { search ended after 'ORIGIN' found }

      end; { if }

    ReadNextString(NextString,NextWord);

    FMain.ProgressBar.Position := Round(CharactersRead / InFileSize * 100);

  end; { while }

    while (NextString <> '//') and not(EOF(InFile)) do
    begin
      ReadNextString(NextString,NextWord);
      FMain.ProgressBar.Position := Round(CharactersRead / InFileSize * 100);
    end;

end; { CompileAdoc }


(********************************************************************)


begin { TCompileThread.Execute }

  { Get variables from FMain }
  Locator := FMain.CBLocator.Text;
  Nucleus := FMain.CBNucleus.Checked;
  Mitochon := FMain.CBMitochon.Checked;
  Chloro := FMain.CBChloro.Checked;
  MultiGenes := FMain.CBMultiGenes.Checked;
  PartialSeqs := FMain.CBPartialSeqs.Checked;
  MarkEnds := FMain.CBMarkEnds.Checked;

  { Get variables from FFiles }
  InFileName := FFiles.EGenBankFile.Text;
  OutFileName := FFiles.ESeqFile.Text;
  LogFileName := FFiles.ELogFile.Text;

  { Find InFile and get size }
  if FindFirst(InFileName, faAnyFile, SearchRec) = 0
    then
      InFileSize := SearchRec.Size;
  FindClose(0);

  { Reset other variables }
  CharactersRead := 0;
  ExaminedDocs := 0;
  SkippedDocs := 0;
  BasesExamined := 0;
  BasesCompiled := 0;

  { Open all files }
  AssignFile(InFile, InFileName);
  Reset(InFile);
  AssignFile(OutFile, OutFileName);
  Rewrite(OutFile);
  AssignFile(LogFile, LogFileName);
  Rewrite(LogFile);

  { Begin writing to LogFile }
  WriteLog(DateToStr(Date) + ' ' + TimeToStr(Time));
  WriteLog('Program: Compile');
  WriteLog('Input File: ' + InFileName + ' (' + FloatToStrF(InFileSize,ffFixed,1,0) + 'bytes)');
  WriteLog('Locator: ' + Locator);
  WriteLog('Output File: ' + OutFileName);
  if Nucleus
    then
      WriteLog('Looking for nucleus origin.');
  if Mitochon
    then
      WriteLog('Looking for mitochondrion origin.');
  if Chloro
    then
      WriteLog('Looking for chloroplast origin.');
  if MultiGenes
    then
      WriteLog('Multiple genes expected per document.')
    else
      WriteLog('Multiple genes NOT expected per document.');
  if PartialSeqs
    then
      WriteLog('Partial sequences are included.')
    else
      WriteLog('Partial sequences are NOT included.');
  if MarkEnds
    then
      WriteLog('Sequence ends are marked.')
    else
      WriteLog('Sequence ends are NOT marked.');

  { Main loop }
  while not EOF(InFile) do
  begin
    CompileAdoc;
    CloseFile(LogFile);
    Append(LogFile);
  end;

  { Finish writing to LogFile }
  WriteLog('');
  WriteLog(FloatToStrF(BasesExamined, ffFixed, 1, 0) + 'bp from ' + IntToStr(ExaminedDocs)
   + ' GenBank documents were examined.');
  WriteLog(FloatToStrF(BasesCompiled, ffFixed, 1, 0) + 'bp (a-z) from ' + IntToStr(ExaminedDocs - SkippedDocs)
   + ' GenBank documents were compiled into ' + OutFileName + '.');

  { Close all files }
  CloseFile(InFile);
  CloseFile(OutFile);
  CloseFile(LogFile);

end; { TCompileThread.Execute }


(******************************************************************************)


end.
